@@ -2,10 +2,12 @@ from __future__ import unicode_literals
 import base64
 import os.path
 import argparse
+import uuid
 
 import py
 from devpi_common.types import cached_property
 from .log import threadlog
+import json
 import devpi_server
 from devpi_common.url import URL
 
@@ -18,7 +20,8 @@ def addoptions(parser):
     web = parser.addgroup("web serving options")
     web.addoption("--host",  type=str,
             default="localhost",
-            help="domain/ip address to listen on")
+            help="domain/ip address to listen on.  Use --host=0.0.0.0 if "
+                 "you want to accept connections from anywhere.")
 
     web.addoption("--port",  type=int,
             default=3141,
@@ -52,7 +55,12 @@ def addoptions(parser):
     deploy.addoption("--version", action="store_true",
             help="show devpi_version (%s)" % devpi_server.__version__)
 
-    deploy.addoption("--master", action="store", dest="master_url",
+    deploy.addoption("--role", action="store", dest="role", default="auto",
+            choices=["master", "replica", "auto"],
+            help="set role of this instance."
+    )
+
+    deploy.addoption("--master-url", action="store", dest="master_url",
             help="run as a replica of the specified master server",
             default=None)
 
@@ -75,10 +83,6 @@ def addoptions(parser):
                  "validation. If it does not exist, a random secret "
                  "is generated on start up and used subsequently. ")
 
-    deploy.addoption("--upgrade-state", action="store_true",
-            dest="upgrade_state",
-            help="(deprecated, use --export/--import) automatically upgrade server state if possible, creating virtualenvs as needed. ")
-
     deploy.addoption("--export", type=str, metavar="PATH",
             help="export devpi-server database state into PATH. "
                  "This will export all users, indices (except root/pypi),"
@@ -102,6 +106,18 @@ def addoptions(parser):
                  "$DEVPI_SERVERDIR is used if it exists, "
                  "otherwise the default is '~/.devpi/server'")
 
+    deploy.addoption("--restrict-modify", type=str, metavar="SPEC",
+            action="store", default=None,
+            help="specify which users/groups may create other users and their "
+                 "indices. Multiple users and groups are separated by commas. "
+                 "Groups need to be prefixed with a colon like this: ':group'. "
+                 "By default anonymous users can create users and "
+                 "then create indices themself, but not modify other users "
+                 "and their indices. The root user can do anything. When this "
+                 "option is set, only the specified users/groups can create "
+                 "and modify users and indices. You have to add root "
+                 "explicitely if wanted.")
+
     bg = parser.addgroup("background server")
     bg.addoption("--start", action="store_true",
             help="start the background devpi-server")
@@ -236,12 +252,26 @@ class PluginManager:
                                   stage=stage,
                                   name2serials=name2serials)
 
+    def devpiserver_auth_user(self, userdict, username, password):
+        """Needs to validate the username and password.
+
+        A dict must be returned with a key "status" with one of the following
+        values:
+            "ok" - authentication succeeded
+            "unknown" - no matching user, other plugins are tried
+            "reject" - invalid password, authentication stops
+
+        Optionally the plugin can return a list of group names the user is
+        member of using the "groups" key of the result dict.
+        """
+        return self._call_plugins("devpiserver_auth_user",
+                                  userdict=userdict,
+                                  username=username, password=password)
+
 
 class Config:
     def __init__(self, args, hook):
         self.args = args
-        if self.args.master_url:
-            self.master_url = URL(self.args.master_url)
         serverdir = args.serverdir
         if serverdir is None:
             serverdir = get_default_serverdir()
@@ -253,8 +283,75 @@ class Config:
             self.secretfile = py.path.local(
                     os.path.expanduser(args.secretfile))
 
+        self.path_nodeinfo = self.serverdir.join(".nodeinfo")
+        self._determine_roles()
+        self._determine_uuid()
+        self.write_nodeinfo()
         self.hook = hook
 
+    def _determine_uuid(self):
+        if "uuid" not in self.nodeinfo:
+            uuid_hex = uuid.uuid4().hex
+            self.nodeinfo["uuid"] = uuid_hex
+            threadlog.info("generated uuid: %s", uuid_hex)
+
+    @property
+    def role(self):
+        return self.nodeinfo["role"]
+
+    def set_uuid(self, uuid):
+        # called when importing state
+        self.nodeinfo["uuid"] = uuid
+        self.write_nodeinfo()
+
+    def set_master_uuid(self, uuid):
+        assert self.role != "master", "cannot set master uuid for master"
+        existing = self.nodeinfo.get("master-uuid")
+        if existing and existing != uuid:
+            raise ValueError("already have master id %r, got %r" % (
+                             existing, uuid))
+        self.nodeinfo["master-uuid"] = uuid
+        self.write_nodeinfo()
+
+    def get_master_uuid(self):
+        if self.role == "master":
+            return self.nodeinfo["uuid"]
+        return self.nodeinfo.get("master-uuid")
+
+    @cached_property
+    def nodeinfo(self):
+        if self.path_nodeinfo.exists():
+            return json.loads(self.path_nodeinfo.read("r"))
+        return {}
+
+    def write_nodeinfo(self):
+        self.path_nodeinfo.dirpath().ensure(dir=1)
+        self.path_nodeinfo.write(json.dumps(self.nodeinfo, indent=2))
+        threadlog.info("wrote nodeinfo to: %s", self.path_nodeinfo)
+
+    def _determine_roles(self):
+        from .main import fatal
+        args = self.args
+        old_role = self.nodeinfo.get("role")
+        if args.master_url:
+            self.master_url = URL(args.master_url)
+            if old_role == "master":
+                fatal("cannot run as replica, was previously run as master")
+            if args.role == "master":
+                fatal("option conflict: --role=master and --master-url")
+            role = "replica"
+            self.nodeinfo["masterurl"] = self.master_url.url
+        else:
+            if args.role == "replica":
+                fatal("need to specify --master-url to run as replica")
+            role = "master"
+        if role == "master" and old_role == "replica" and \
+           args.role != "master":
+            fatal("need to specify --role=master to run previous replica "
+                  "as a master")
+        self.nodeinfo["role"] = role
+        return
+
     @cached_property
     def secret(self):
         if not self.secretfile.check():
