@@ -3,67 +3,13 @@ import sys
 import os
 import json
 import py
-import subprocess
 import logging
 from devpi_common.validation import normalize_name
 from devpi_common.metadata import Version, splitbasename, BasenameMeta
-from devpi_server.main import fatal, server_version
-from devpi_server.venv import create_server_venv
+from devpi_server.main import fatal
 import devpi_server
 
 
-def do_upgrade(xom):
-    tw = py.io.TerminalWriter()
-    serverdir = xom.config.serverdir
-    exportdir = serverdir + "-export"
-    if exportdir.check():
-        tw.line("removing exportdir: %s" % exportdir)
-        exportdir.remove()
-    newdir = serverdir + "-import"
-    script = sys.argv[0]
-    def rel(p):
-        return py.path.local().bestrelpath(p)
-    state_version = xom.get_state_version()
-    tw.sep("-", "exporting to %s" % rel(exportdir))
-    if Version(state_version) > Version(server_version):
-        fatal("%s has state version %s which is newer than what we "
-              "can handle (%s)" %(
-                xom.config.serverdir, state_version, server_version))
-    elif Version(state_version) < Version(server_version):
-        tw.line("creating server venv %s ..." % state_version)
-        tmpdir = py.path.local.make_numbered_dir("devpi-upgrade")
-        venv = create_server_venv(tw, state_version, tmpdir)
-        tw.line("server venv %s created: %s" % (server_version, tmpdir))
-        venv.check_call(["devpi-server",
-                         "--serverdir", str(serverdir),
-                         "--export", str(exportdir)])
-    else:
-        #tw.sep("-", "creating venv" % rel(exportdir))
-        subprocess.check_call([sys.executable, script,
-                               "--serverdir", str(serverdir),
-                               "--export", str(exportdir)])
-    tw.sep("-", "importing from %s" % rel(exportdir))
-    tw.line("importing into server version: %s" % server_version)
-    subprocess.check_call([sys.executable, script,
-                           "--serverdir", str(newdir),
-                           "--import", str(exportdir)])
-    tw.sep("-", "replacing serverstate")
-    backup_dir = serverdir + "-backup"
-    if backup_dir.check():
-        tw.line("backup dir exists, not creating backup", bold=True)
-    else:
-        tw.line("moving serverstate to backupdir: %s" % (backup_dir), bold=True)
-        serverdir.move(backup_dir)
-    if serverdir.check():
-        tw.line("removing serverstate: %s" % (serverdir))
-        serverdir.remove()
-    tw.line("copying new serverstate to serverdir", bold=True)
-    newdir.move(serverdir)
-    serverdir.join(".serverversion").read()
-    tw.line("cleanup: removing exportdir: %s" % exportdir)
-    tw.line("have fun serving the new state :)")
-    exportdir.remove()
-
 def do_export(path, xom):
     path = py.path.local(path)
     tw = py.io.TerminalWriter()
@@ -125,6 +71,7 @@ class Exporter:
         self.export["pythonversion"] = list(sys.version_info)
         self.export["devpi_server"] = devpi_server.__version__
         self.export["secret"] = self.config.secret
+        self.export["uuid"] = self.xom.config.get_master_uuid()
         self.compute_global_projectname_normalization()
         for user in self.xom.model.get_userlist():
             userdir = path.join(user.name)
@@ -234,14 +181,11 @@ class IndexDump:
                 self.basedir.join(linkstore.projectname, entry.basename))
             self.add_filedesc("releasefile", linkstore.projectname, relpath,
                                version=linkstore.version,
-                               entrymapping=entry.meta.copy())
+                               entrymapping=entry.meta.copy(),
+                               log=link.get_logs())
 
     def dump_toxresults(self, linkstore):
         for tox_link in linkstore.get_links(rel="toxresult"):
-            if not tox_link.entry.file_exists():
-                self.exporter.warn("skipping missing toxresult file: %r" %
-                                   tox_link)
-                continue
             reflink = linkstore.stage.get_link_from_entrypath(tox_link.for_entrypath)
             relpath = self.exporter.write_file(
                 content=tox_link.entry.file_get_content(),
@@ -252,7 +196,8 @@ class IndexDump:
                               projectname=linkstore.projectname,
                               relpath=relpath,
                               version=linkstore.version,
-                              for_entrypath=reflink.entrypath)
+                              for_entrypath=reflink.entrypath,
+                              log=tox_link.get_logs())
 
     def add_filedesc(self, type, projectname, relpath, **kw):
         assert self.exporter.basepath.join(relpath).check()
@@ -291,6 +236,9 @@ class Importer:
         self.dumpversion = self.import_data["dumpversion"]
         if self.dumpversion not in ("1", "2"):
             fatal("incompatible dumpversion: %r" %(self.dumpversion,))
+        uuid = self.import_data.get("uuid")
+        if uuid is not None:
+            self.xom.config.set_uuid(uuid)
         self.import_users = self.import_data["users"]
         self.import_indexes = self.import_data["indexes"]
         self.xom.config.secret = secret = self.import_data["secret"]
@@ -337,8 +285,8 @@ class Importer:
                 for version, versiondata in versions.items():
                     with self.xom.keyfs.transaction(write=True):
                         assert "+elinks" not in versiondata
-                        if '+doczip' in versiondata:
-                            del versiondata['+doczip']
+                        versiondata.pop('+doczip', None)
+                        versiondata.pop(':action', None)
                         assert not any(True for x in versiondata if x.startswith('+'))
                         if not versiondata.get("version"):
                             name = versiondata["name"]
@@ -374,22 +322,27 @@ class Importer:
             else:
                 version = filedesc["version"]
 
-            entry = stage.store_releasefile(projectname, version,
-                                            p.basename, p.read("rb"),
-                                            last_modified=mapping["last_modified"])
-            assert entry.md5 == mapping["md5"]
-            self.import_pre2_toxresults(stage, entry)
+            link = stage.store_releasefile(projectname, version,
+                                           p.basename, p.read("rb"),
+                                           last_modified=mapping["last_modified"])
+            assert link.entry.md5 == mapping["md5"]
+            self.import_pre2_toxresults(stage, link.entry)
         elif filedesc["type"] == "doczip":
             basename = os.path.basename(rel)
             name, version, suffix = splitbasename(basename)
-            stage.store_doczip(name, version, p.read("rb"))
+            link = stage.store_doczip(name, version, p.read("rb"))
         elif filedesc["type"] == "toxresult":
             linkstore = stage.get_linkstore_perstage(filedesc["projectname"],
                                            filedesc["version"])
             link, = linkstore.get_links(entrypath=filedesc["for_entrypath"])
-            stage.store_toxresult(link, json.loads(p.read("rb").decode("utf8")))
+            link = stage.store_toxresult(link, json.loads(p.read("rb").decode("utf8")))
         else:
             fatal("unknown file type: %s" % (type,))
+        history_log = filedesc.get('log')
+        if history_log is None:
+            link.add_log('upload', '<import>', dst=stage.name)
+        else:
+            link.add_logs(history_log)
 
     def import_pre2_toxresults(self, stage, releasefile_entry):
         # pre 2.0 export structure (called "attachments")
@@ -405,8 +358,9 @@ class Importer:
             attachment_data = attachment_data.decode('utf-8')
             toxresultdata = json.loads(attachment_data)
             self.tw.line("importing pre-2.0 test  results %s/%s" %(md5, type))
-            link = stage.store_toxresult(releasefile_link, toxresultdata)
-            self.tw.line("imported %s" % link.entrypath)
+            tox_link = stage.store_toxresult(releasefile_link, toxresultdata)
+            tox_link.add_log('upload', '<import>')
+            self.tw.line("imported %s" % tox_link.entrypath)
 
 
 class IndexTree:
