@@ -13,6 +13,7 @@ from devpi_common.request import new_requests_session
 from .config import PluginManager
 from .config import parseoptions, load_setuptools_entrypoints
 from .log import configure_logging, threadlog
+from .model import BaseStage
 from . import extpypi, replica, mythread
 from . import __version__ as server_version
 
@@ -27,18 +28,17 @@ def fatal(msg):
 
 def check_compatible_version(xom):
     args = xom.config.args
-    if args.upgrade_state or args.export:
+    if args.export:
         return
     state_version = xom.get_state_version()
     if server_version != state_version:
-        state_ver = map(int, state_version.split(".")[:2])
-        server_ver = map(int, server_version.split(".")[:2])
+        state_ver = list(map(int, state_version.split(".")[:2]))
+        server_ver = list(map(int, server_version.split(".")[:2]))
         if state_ver != server_ver:
             fatal("Incompatible state: server %s cannot run serverdir "
-                  "%s created by %s. "
+                  "%s created by %s.\n"
                   "Use --export from older version, then --import with newer "
-                  "version.  Or try --upgrade-state for in-place upgrades."
-                  " But do a backup first :)"
+                  "version."
                   %(server_version, xom.config.serverdir, state_version))
 
         xom.set_state_version(server_version)
@@ -122,6 +122,7 @@ def wsgi_run(xom, app):
     log = xom.log
     log.info("devpi-server version: %s", server_version)
     log.info("serverdir: %s" % xom.config.serverdir)
+    log.info("uuid: %s" % xom.config.nodeinfo["uuid"])
     hostaddr = "http://%s:%s" % (host, port)
     log.info("serving at url: %s", hostaddr)
     log.info("bug tracker: https://bitbucket.org/hpk42/devpi/issues")
@@ -136,7 +137,6 @@ def wsgi_run(xom, app):
         pass
     return 0
 
-
 class XOM:
     class Exiting(SystemExit):
         pass
@@ -149,16 +149,17 @@ class XOM:
         if httpget is not None:
             self.httpget = httpget
         sdir = config.serverdir
-        if not (sdir.exists() and sdir.listdir()):
+        if not (sdir.exists() and len(sdir.listdir()) >= 2):
             self.set_state_version(server_version)
         self.log = threadlog
+        self.polling_replicas = {}
 
     def get_state_version(self):
         versionfile = self.config.serverdir.join(".serverversion")
         if not versionfile.exists():
-            fatal("serverdir %s is unversioned, please use depvi-server-1.1 "
-              "with the --upgrade-state option to upgrade from versions "
-              "prior to 1.1\n" % self.config.serverdir)
+            fatal("serverdir %s is unversioned, you may try use "
+              "depvi-server-1.1 with the --upgrade-state option to "
+              "upgrade from versions prior to 1.1\n" % self.config.serverdir)
         return versionfile.read()
 
     def set_state_version(self, version):
@@ -175,10 +176,6 @@ class XOM:
     def main(self):
         xom = self
         args = xom.config.args
-        if args.upgrade_state:
-            from devpi_server.importexport import do_upgrade
-            return do_upgrade(xom)
-
         # need to initialize the pypi mirror state before importing
         # because importing may need pypi mirroring state
         if xom.is_replica():
@@ -301,6 +298,7 @@ class XOM:
 
         pyramid_config.add_route("/+changelog/{serial}",
                                  "/+changelog/{serial}")
+        pyramid_config.add_route("/+status", "/+status")
         pyramid_config.add_route("/root/pypi/+name2serials",
                                  "/root/pypi/+name2serials")
         pyramid_config.add_route("/+api", "/+api", accept="application/json")
@@ -334,6 +332,9 @@ class XOM:
         pyramid_config.add_tween("devpi_server.views.tween_keyfs_transaction",
             under="devpi_server.views.tween_request_logging"
         )
+        pyramid_config.add_request_method(get_remote_ip)
+        pyramid_config.add_request_method(stage_url)
+        pyramid_config.add_request_method(simpleindex_url)
 
         # overwrite route_url method with our own
         pyramid_config.add_request_method(route_url)
@@ -358,13 +359,41 @@ class XOM:
     def is_replica(self):
         return bool(self.config.args.master_url)
 
-
 class FatalResponse:
     status_code = -1
 
     def __init__(self, excinfo=None):
         self.excinfo = excinfo
 
+def get_remote_ip(request):
+    return request.headers.get("X-REAL-IP", request.client_addr)
+
+
+def stage_from_args(model, *args):
+    if len(args) not in (1, 2):
+        raise TypeError("stage_url() takes 1 or 2 arguments (%s given)" % len(args))
+    if len(args) == 1:
+        if isinstance(args[0], BaseStage):
+            return args[0]
+    return model.getstage(*args)
+
+
+def stage_url(request, *args):
+    model = request.registry['xom'].model
+    stage = stage_from_args(model, *args)
+    if stage is not None:
+        return request.route_url(
+            "/{user}/{index}", user=stage.username, index=stage.index)
+
+
+def simpleindex_url(request, *args):
+    model = request.registry['xom'].model
+    stage = stage_from_args(model, *args)
+    if stage is not None:
+        return request.route_url(
+            "/{user}/{index}/+simple/", user=stage.username, index=stage.index)
+
+
 def set_default_indexes(model):
     root_user = model.get_user("root")
     if not root_user:
