@@ -6,6 +6,7 @@ import json
 from devpi_common.metadata import sorted_sameproject_links, get_latest_version
 from devpi_common.validation import validate_metadata, normalize_name
 from devpi_common.types import ensure_unicode, cached_property
+from time import gmtime
 from .auth import crypt_password, verify_password
 from .filestore import FileEntry, split_md5
 from .log import threadlog, thread_current_log
@@ -60,6 +61,10 @@ class MissesRegistration(ModelException):
     """ A prior registration or release metadata is required. """
 
 
+class NonVolatile(ModelException):
+    """ A release is overwritten on a non volatile index. """
+
+
 class RootModel:
     def __init__(self, xom):
         self.xom = xom
@@ -237,13 +242,17 @@ class BaseStage:
     NotFound = NotFound
     UpstreamError = UpstreamError
     MissesRegistration = MissesRegistration
+    NonVolatile = NonVolatile
 
     def get_linkstore_perstage(self, name, version):
         return LinkStore(self, name, version)
 
     def get_link_from_entrypath(self, entrypath):
         entry = self.xom.filestore.get_file_entry(entrypath)
-        linkstore = self.get_linkstore_perstage(entry.projectname, entry.version)
+        if entry.projectname is None:
+            return None
+        linkstore = self.get_linkstore_perstage(entry.projectname,
+                                                entry.version)
         links = linkstore.get_links(entrypath=entrypath)
         assert len(links) < 2
         return links[0] if links else None
@@ -382,6 +391,7 @@ class PrivateStage(BaseStage):
         self.model = xom.model
         self.keyfs = xom.keyfs
         self.user = self.model.get_user(user)
+        self.username = self.user.name
         self.index = index
         self.name = user + "/" + index
         self.ixconfig = ixconfig
@@ -460,12 +470,12 @@ class PrivateStage(BaseStage):
         self._set_versiondata(metadata)
 
     def key_projversions(self, name):
-        name = self.get_projectname_perstage(name) or name
+        name = normalize_name(name)
         return self.keyfs.PROJVERSIONS(
             user=self.user.name, index=self.index, name=name)
 
     def key_projversion(self, name, version):
-        name = self.get_projectname_perstage(name) or name
+        name = normalize_name(name)
         return self.keyfs.PROJVERSION(
             user=self.user.name, index=self.index, name=name, version=version)
 
@@ -551,12 +561,12 @@ class PrivateStage(BaseStage):
                 raise MissesRegistration("%s-%s", name, version)
         threadlog.debug("project name of %r is %r", filename, name)
         linkstore = self.get_linkstore_perstage(name, version)
-        entry = linkstore.create_linked_entry(
+        link = linkstore.create_linked_entry(
                 rel="releasefile",
                 basename=filename,
                 file_content=content,
                 last_modified=last_modified)
-        return entry
+        return link
 
     def store_doczip(self, name, version, content):
         if not version:
@@ -565,12 +575,12 @@ class PrivateStage(BaseStage):
                            name, version)
         basename = "%s-%s.doc.zip" % (name, version)
         linkstore = self.get_linkstore_perstage(name, version)
-        entry = linkstore.create_linked_entry(
+        link = linkstore.create_linked_entry(
                 rel="doczip",
                 basename=basename,
                 file_content=content,
         )
-        return entry
+        return link
 
     def get_doczip(self, name, version):
         """ get documentation zip as an open file
@@ -582,7 +592,6 @@ class PrivateStage(BaseStage):
             return links[0].entry.file_get_content()
 
 
-
 class ELink:
     """ model Link using entrypathes for referencing. """
     def __init__(self, filestore, linkdict, projectname, version):
@@ -607,6 +616,15 @@ class ELink:
     def entry(self):
         return self.filestore.get_file_entry(self.entrypath)
 
+    def add_log(self, what, who, **kw):
+        self._log.append(dict(what=what, who=who, when=gmtime()[:6], **kw))
+
+    def add_logs(self, logs):
+        self._log.extend(logs)
+
+    def get_logs(self):
+        return list(self._log)
+
 
 class LinkStore:
     def __init__(self, stage, projectname, version):
@@ -619,17 +637,26 @@ class LinkStore:
             raise MissesRegistration("%s-%s on stage %s",
                                      projectname, version, stage.name)
 
+    def get_file_entry(self, relpath):
+        return self.filestore.get_file_entry(relpath)
+
     def create_linked_entry(self, rel, basename, file_content, last_modified=None):
         assert isinstance(file_content, bytes)
+        overwrite = None
         for link in self.get_links(rel=rel, basename=basename):
             if not self.stage.ixconfig.get("volatile"):
-                return 409
+                raise NonVolatile("rel=%s basename=%s on stage %s" % (
+                    rel, basename, self.stage.name))
+            assert overwrite is None
+            overwrite = sum(x.get('count', 0) for x in link.get_logs() if x.get('what') == 'overwrite')
             self.remove_links(rel=rel, basename=basename)
         file_entry = self._create_file_entry(basename, file_content)
         if last_modified is not None:
             file_entry.last_modified = last_modified
-        self._add_link_to_file_entry(rel, file_entry)
-        return file_entry
+        link = self._add_link_to_file_entry(rel, file_entry)
+        if overwrite is not None:
+            link.add_log('overwrite', None, count=overwrite + 1)
+        return link
 
     def new_reflink(self, rel, file_content, for_entrypath):
         if isinstance(for_entrypath, ELink):
@@ -698,7 +725,7 @@ class LinkStore:
             relextra["for_entrypath"] = for_entrypath
         linkdicts = self._get_inplace_linkdicts()
         new_linkdict = dict(rel=rel, entrypath=file_entry.relpath,
-                            md5=file_entry.md5, **relextra)
+                            md5=file_entry.md5, _log=[], **relextra)
         linkdicts.append(new_linkdict)
         threadlog.info("added %r link %s", rel, file_entry.relpath)
         self._mark_dirty()
