@@ -1,6 +1,8 @@
 import os
 import py
 import hashlib
+import contextlib
+import time
 from pyramid.httpexceptions import HTTPNotFound, HTTPAccepted
 from pyramid.view import view_config
 from pyramid.response import Response
@@ -8,9 +10,11 @@ from webob.headers import EnvironHeaders, ResponseHeaders
 
 from .keyfs import load, loads, dump, get_write_file_ensure_dir
 from .log import thread_push_log, threadlog
-from .views import is_mutating_http_method
+from .views import is_mutating_http_method, H_MASTER_UUID
 from .model import UpstreamError
 
+H_REPLICA_UUID = str("X-DEVPI-REPLICA-UUID")
+H_REPLICA_OUTSIDE_URL = str("X-DEVPI-REPLICA-OUTSIDE-URL")
 
 class MasterChangelogRequest:
     MAX_REPLICA_BLOCK_TIME = 30.0
@@ -20,6 +24,27 @@ class MasterChangelogRequest:
         self.request = request
         self.xom = request.registry["xom"]
 
+    @contextlib.contextmanager
+    def update_replica_status(self, serial):
+        headers = self.request.headers
+        uuid = headers.get(H_REPLICA_UUID)
+        if uuid:
+            polling_replicas = self.xom.polling_replicas
+            polling_replicas[uuid] = {
+                "remote-ip": self.request.get_remote_ip(),
+                "serial": serial,
+                "in-request": True,
+                "last-request": time.time(),
+                "outside-url": headers.get(H_REPLICA_OUTSIDE_URL),
+            }
+            try:
+                yield
+            finally:
+                polling_replicas[uuid]["last-request"] = time.time()
+                polling_replicas[uuid]["in-request"] = False
+        else:  # just a regular request
+            yield
+
     @view_config(route_name="/+changelog/{serial}")
     def get_changes(self):
         # this method is called from all replica servers
@@ -34,24 +59,26 @@ class MasterChangelogRequest:
         # if no commits happen.
 
         serial = self.request.matchdict["serial"]
-        keyfs = self.xom.keyfs
-        if serial.lower() == "nop":
-            raw_entry = b""
-        else:
-            try:
-                serial = int(serial)
-            except ValueError:
-                raise HTTPNotFound("serial needs to be int")
-            raw_entry = keyfs._fs.get_raw_changelog_entry(serial)
-            if not raw_entry:
-                raw_entry = self._wait_for_entry(serial)
 
-        devpi_serial = keyfs.get_current_serial()
-        r = Response(body=raw_entry, status=200, headers={
-            str("Content-Type"): str("application/octet-stream"),
-            str("X-DEVPI-SERIAL"): str(devpi_serial),
-        })
-        return r
+        with self.update_replica_status(serial):
+            keyfs = self.xom.keyfs
+            if serial.lower() == "nop":
+                raw_entry = b""
+            else:
+                try:
+                    serial = int(serial)
+                except ValueError:
+                    raise HTTPNotFound("serial needs to be int")
+                raw_entry = keyfs._fs.get_raw_changelog_entry(serial)
+                if not raw_entry:
+                    raw_entry = self._wait_for_entry(serial)
+
+            devpi_serial = keyfs.get_current_serial()
+            r = Response(body=raw_entry, status=200, headers={
+                str("Content-Type"): str("application/octet-stream"),
+                str("X-DEVPI-SERIAL"): str(devpi_serial),
+            })
+            return r
 
     def _wait_for_entry(self, serial):
         max_wakeups = self.MAX_REPLICA_BLOCK_TIME / self.WAKEUP_INTERVAL
@@ -84,6 +111,8 @@ class MasterChangelogRequest:
 
 
 class ReplicaThread:
+    ERROR_SLEEP = 50
+
     def __init__(self, xom):
         self.xom = xom
         self.master_url = xom.config.master_url
@@ -98,32 +127,55 @@ class ReplicaThread:
         keyfs = self.xom.keyfs
         for key in (keyfs.STAGEFILE, keyfs.PYPIFILE_NOMD5):
             keyfs.subscribe_on_import(key, ImportFileReplica(self.xom))
+        config = self.xom.config
+        nodeinfo = config.nodeinfo
         while 1:
             self.thread.exit_if_shutdown()
             serial = keyfs.get_next_serial()
             url = self.master_url.joinpath("+changelog", str(serial)).url
             log.info("fetching %s", url)
             try:
-                r = session.get(url, stream=True)
-            except session.Errors:
-                log.exception("error fetching %s", url)
+                r = session.get(url, headers={
+                    H_REPLICA_UUID: nodeinfo["uuid"],
+                    H_REPLICA_OUTSIDE_URL: config.args.outside_url,
+                })
+            except session.Errors as e:
+                log.error("error fetching %s: %s", url, str(e))
             else:
+                # we check that the remote instance
+                # has the same UUID we saw last time
+                master_uuid = config.get_master_uuid()
+                remote_master_uuid = r.headers.get(H_MASTER_UUID)
+                if not remote_master_uuid:
+                    log.error("remote provides no %r header, running "
+                              "<devpi-server-2.1?"
+                              " headers were: %s", H_MASTER_UUID, r.headers)
+                    self.thread.sleep(self.ERROR_SLEEP)
+                    continue
+                if master_uuid and remote_master_uuid != master_uuid:
+                    log.error("master UUID %r does not match "
+                              "locally recorded master UUID %r",
+                              remote_master_uuid, master_uuid)
+                    self.thread.sleep(self.ERROR_SLEEP)
+                    continue
                 if r.status_code == 200:
                     try:
-                        entry = loads(r.raw.read())
-                    except Exception:
-                        log.exception("could not read answer %s", url)
-                    else:
-                        changes, rel_renames = entry
+                        changes, rel_renames = loads(r.content)
                         keyfs.import_changes(serial, changes)
+                    except Exception:
+                        log.exception("could not process: %s", r.url)
+                    else:
                         serial += 1
+                        # we successfully received data so let's
+                        # record the master_uuid for future consistency checks
+                        if not master_uuid:
+                            self.xom.config.set_master_uuid(remote_master_uuid)
                         continue
                 elif r.status_code == 202:
                     log.debug("%s: trying again %s\n", r.status_code, url)
                     continue
                 else:
-                    log.debug("%s: failed fetching %s\n%s",
-                              r.status_code, url, getattr(r, 'text', ''))
+                    log.debug("%s: failed fetching %s", r.status_code, url)
             # we got an error, let's wait a bit
             self.thread.sleep(5.0)
 
@@ -215,24 +267,27 @@ def clean_response_headers(response):
     return headers
 
 
-def proxy_write_to_master(xom, request):
-    """ relay modifying http requests to master and wait until
-    the change is replicated back.
-    """
+def proxy_request_to_master(xom, request):
     master_url = xom.config.master_url
     url = master_url.joinpath(request.path).url
     http = xom._httpsession
     with threadlog.around("info", "relaying: %s %s", request.method, url):
         try:
-            r = http.request(request.method, url,
-                             data=request.body,
-                             headers=clean_request_headers(request),
-                             allow_redirects=False)
+            return http.request(request.method, url,
+                                data=request.body,
+                                headers=clean_request_headers(request),
+                                allow_redirects=False)
         except http.Errors as e:
             raise UpstreamError("proxy-write-to-master %s: %s" % (url, e))
+
+
+def proxy_write_to_master(xom, request):
+    """ relay modifying http requests to master and wait until
+    the change is replicated back.
+    """
+    r = proxy_request_to_master(xom, request)
     #threadlog.debug("relay status_code: %s", r.status_code)
     #threadlog.debug("relay headers: %s", r.headers)
-
     if r.status_code < 400:
         commit_serial = int(r.headers["X-DEVPI-SERIAL"])
         xom.keyfs.notifier.wait_tx_serial(commit_serial)
@@ -243,7 +298,7 @@ def proxy_write_to_master(xom, request):
         master_location = r.headers["location"]
         outside_url = request.application_url
         headers[str("location")] = str(
-            master_location.replace(master_url.url, outside_url))
+            master_location.replace(xom.config.master_url.url, outside_url))
     return Response(status="%s %s" %(r.status_code, r.reason),
                     body=r.content,
                     headers=headers)
@@ -272,14 +327,26 @@ class ImportFileReplica:
         url = self.xom.config.master_url.joinpath(relpath).url
         r = self.xom.httpget(url, allow_redirects=True)
         if r.status_code != 200:
-            threadlog.error("got %s from upstream", r.status_code)
-            return
+            raise FileReplicationError(r)
         remote_md5 = hashlib.md5(r.content).hexdigest()
         if entry.md5 and entry.md5 != remote_md5:
-            threadlog.error("%s: remote has md5 %s, expected %s",
-                            url, remote_md5, entry.md5)
-        else:
-            tmppath = entry._filepath + "-tmp"
-            with get_write_file_ensure_dir(tmppath) as f:
-                f.write(r.content)
-            fswriter.record_rename_file(tmppath, entry._filepath)
+            raise FileReplicationError(r, "remote has md5 %s, expected %s" %(
+                                       remote_md5, entry.md5))
+        tmppath = entry._filepath + "-tmp"
+        with get_write_file_ensure_dir(tmppath) as f:
+            f.write(r.content)
+        fswriter.record_rename_file(tmppath, entry._filepath)
+
+
+class FileReplicationError(Exception):
+    """ raised when replicating a file from the master failed. """
+    def __init__(self, response, message=None):
+        self.url = response.url
+        self.status_code = response.status_code
+        self.message = message or "failed"
+
+    def __str__(self):
+        return "FileReplicationError with %s, code=%s, message=%s" %(
+               self.url, self.status_code, self.message)
+
+
