@@ -1,5 +1,6 @@
 from __future__ import unicode_literals
 
+import os
 import py
 from py.xml import html
 from devpi_common.types import ensure_unicode
@@ -30,6 +31,8 @@ server_version = devpi_server.__version__
 
 MAXDOCZIPSIZE = 30 * 1024 * 1024    # 30MB
 
+H_MASTER_UUID = str("X-DEVPI-MASTER-UUID")
+
 
 API_VERSION = "2"
 
@@ -105,6 +108,8 @@ class OutsideURLMiddleware(object):
         if not outside_url:
             outside_url = environ.get('HTTP_X_OUTSIDE_URL')
         if outside_url:
+            # XXX memoize it for later access from replica thread
+            # self.xom.current_outside_url = outside_url
             outside_url = urlparse.urlparse(outside_url)
             environ['wsgi.url_scheme'] = outside_url.scheme
             environ['HTTP_HOST'] = outside_url.netloc
@@ -126,6 +131,8 @@ def tween_request_logging(handler, registry):
     req_count = itertools.count()
     from time import time
 
+    nodeinfo = registry["xom"].config.nodeinfo
+
     def request_log_handler(request):
         tag = "[req%s]" %(next(req_count))
         log = thread_push_log(tag)
@@ -137,6 +144,10 @@ def tween_request_logging(handler, registry):
         rheaders = response.headers
         serial = rheaders.get("X-DEVPI-SERIAL")
         rheaders.update(meta_headers)
+        uuid, master_uuid = make_uuid_headers(nodeinfo)
+        rheaders[str("X-DEVPI-UUID")] = str(uuid)
+        rheaders[H_MASTER_UUID] = str(master_uuid)
+
         log.debug("%s %.3fs serial=%s length=%s type=%s",
                   response.status_code,
                   duration,
@@ -149,6 +160,13 @@ def tween_request_logging(handler, registry):
     return request_log_handler
 
 
+def make_uuid_headers(nodeinfo):
+    uuid = master_uuid = nodeinfo.get("uuid")
+    if uuid is not None and nodeinfo["role"] == "replica":
+        master_uuid = nodeinfo.get("master-uuid", "")
+    return uuid, master_uuid
+
+
 def tween_keyfs_transaction(handler, registry):
     keyfs = registry["xom"].keyfs
     is_replica = registry["xom"].is_replica()
@@ -170,6 +188,37 @@ def set_header_devpi_serial(headers, serial):
 def is_mutating_http_method(method):
     return method in ("PUT", "POST", "PATCH", "DELETE", "PUSH")
 
+class StatusView:
+    def __init__(self, request):
+        self.request = request
+        self.xom = request.registry["xom"]
+
+    @view_config(route_name="/+status")
+    def status(self):
+        config = self.xom.config
+
+        status = {
+            "serverdir": str(config.serverdir),
+            "uuid": self.xom.config.nodeinfo["uuid"],
+            "versioninfo":
+                    dict(self.request.registry["devpi_version_info"]),
+            "server-code": os.path.dirname(devpi_server.__file__),
+            "host": config.args.host,
+            "port": config.args.port,
+            "outside-url": config.args.outside_url,
+            "serial": self.xom.keyfs.get_current_serial(),
+            "event-serial": self.xom.keyfs.notifier.read_event_serial(),
+        }
+        master_url = config.args.master_url
+        if master_url:
+            status["role"] = "REPLICA"
+            status["master-url"] = master_url
+            status["master-uuid"] = config.nodeinfo.get("master-uuid")
+        else:
+            status["role"] = "MASTER"
+        status["polling_replicas"] = self.xom.polling_replicas
+        apireturn(200, type="status", result=status)
+
 
 class PyPIView:
     def __init__(self, request):
@@ -182,6 +231,13 @@ class PyPIView:
         self.log = request.log
 
     def get_auth_status(self):
+        if self.xom.is_replica():
+            from .replica import proxy_request_to_master
+            r = proxy_request_to_master(self.xom, self.request)
+            if r.status_code == 200:
+                return r.json()["result"]["authstatus"]
+            threadlog.error("could not obtain master authentication status")
+            return ["fail", "", []]
         # this is accessing some pyramid internals, but they are pretty likely
         # to stay and the alternative was uglier
         policy = self.request._get_authentication_policy()
@@ -207,10 +263,8 @@ class PyPIView:
                 user, index = parts[:2]
                 stage = self.context.getstage(user, index)
                 api.update({
-                    "index": request.route_url(
-                        "/{user}/{index}", user=user, index=index),
-                    "simpleindex": request.route_url(
-                        "/{user}/{index}/+simple/", user=user, index=index)
+                    "index": request.stage_url(stage),
+                    "simpleindex": request.simpleindex_url(stage)
                 })
                 if stage.ixconfig["type"] == "stage":
                     api["pypisubmit"] = request.route_url(
@@ -231,6 +285,8 @@ class PyPIView:
             apireturn(404, message="no release file found at %s" % relpath)
         toxresultdata = getjson(self.request)
         tox_link = stage.store_toxresult(link, toxresultdata)
+        tox_link.add_log(
+            'upload', self.request.authenticated_userid, dst=stage.name)
         apireturn(200, type="toxresultpath",
                   result=tox_link.entrypath)
 
@@ -410,9 +466,6 @@ class PyPIView:
 
         metadata = get_pure_metadata(linkstore.verdata)
 
-        # prepare metadata for submission
-        metadata[":action"] = "submit"
-
         results = []
         targetindex = pushdata.get("targetindex", None)
         if targetindex is not None:  # in-server push
@@ -436,6 +489,8 @@ class PyPIView:
             username = pushdata["username"]
             password = pushdata["password"]
             pypiauth = (username, password)
+            # prepare metadata for submission
+            metadata[":action"] = "submit"
             self.log.info("registering %s-%s to %s", name, version, posturl)
             session = new_requests_session(agent=("server", server_version))
             r = session.post(posturl, data=metadata, auth=pypiauth)
@@ -477,8 +532,18 @@ class PyPIView:
 
     def _push_links(self, links, target_stage, name, version):
         for link in links["releasefile"]:
-            entry = target_stage.store_releasefile(
-                name, version, link.basename, link.entry.file_get_content())
+            new_link = target_stage.store_releasefile(
+                name, version, link.basename, link.entry.file_get_content(),
+                last_modified=link.entry.last_modified)
+            new_link.add_logs(
+                x for x in link.get_logs()
+                if x.get('what') != 'overwrite')
+            new_link.add_log(
+                'push',
+                self.request.authenticated_userid,
+                src=self.context.stage.name,
+                dst=target_stage.name)
+            entry = new_link.entry
             yield (200, "store_releasefile", entry.relpath)
             # also store all dependent tox results
             tstore = target_stage.get_linkstore_perstage(name, version)
@@ -488,10 +553,24 @@ class PyPIView:
                     raw_data = toxlink.entry.file_get_content()
                     data = json.loads(raw_data.decode("utf-8"))
                     tlink = target_stage.store_toxresult(ref_link, data)
+                    tlink.add_logs(
+                        x for x in toxlink.get_logs()
+                        if x.get('what') != 'overwrite')
+                    tlink.add_log(
+                        'push',
+                        self.request.authenticated_userid,
+                        src=self.context.stage.name,
+                        dst=target_stage.name)
                     yield (200, "store_toxresult", tlink.entrypath)
         for link in links["doczip"]:
             doczip = link.entry.file_get_content()
-            target_stage.store_doczip(name, version, doczip)
+            new_link = target_stage.store_doczip(name, version, doczip)
+            new_link.add_logs(link.get_logs())
+            new_link.add_log(
+                'push',
+                self.request.authenticated_userid,
+                src=self.context.stage.name,
+                dst=target_stage.name)
             yield (200, "store_doczip", name, version,
                    "->", target_stage.name)
             break # we only can have one doczip for now
@@ -539,14 +618,19 @@ class PyPIView:
                     metadata = stage.get_versiondata(projectname, version)
                     if not metadata:
                         abort_submit(400, "could not process form metadata")
-                res = stage.store_releasefile(projectname, version,
-                                              content.filename, content.file.read())
-                if res == 409:
+                try:
+                    link = stage.store_releasefile(
+                        projectname, version,
+                        content.filename, content.file.read())
+                except stage.NonVolatile:
                     abort_submit(409, "%s already exists in non-volatile index" % (
                          content.filename,))
+                link.add_log(
+                    'upload', request.authenticated_userid, dst=stage.name)
                 jenkinurl = stage.ixconfig["uploadtrigger_jenkins"]
                 if jenkinurl:
-                    jenkinurl = jenkinurl.format(pkgname=name)
+                    jenkinurl = jenkinurl.format(pkgname=name,
+                                                 pkgversion=version)
                     if trigger_jenkins(request, stage, jenkinurl, name) == -1:
                         abort_submit(200,
                             "OK, but couldn't trigger jenkins at %s" %
@@ -556,7 +640,12 @@ class PyPIView:
                 if len(doczip) > MAXDOCZIPSIZE:
                     abort_submit(413, "zipfile size %d too large, max=%s"
                                    % (len(doczip), MAXDOCZIPSIZE))
-                stage.store_doczip(name, version, doczip)
+                try:
+                    link = stage.store_doczip(name, version, doczip)
+                except stage.MissesRegistration:
+                    apireturn(400, "%s-%s is not registered" %(name, version))
+                link.add_log(
+                    'upload', request.authenticated_userid, dst=stage.name)
         else:
             abort_submit(400, "action %r not supported" % action)
         return Response("")
@@ -657,6 +746,8 @@ class PyPIView:
                 if for_entrypath is not None:
                     linkdict["for_href"] = \
                         url_for_entrypath(self.request, for_entrypath)
+                if "_log" in linkdict:
+                    linkdict["log"] = list(linkdict.pop("_log"))
                 links.append(linkdict)
         shadowing = view_verdata.pop("+shadowing", None)
         if shadowing:
@@ -756,7 +847,9 @@ class PyPIView:
                                                       password=password))
         apireturn(200, "user updated")
 
-    @view_config(route_name="/{user}", request_method="PUT")
+    @view_config(
+        route_name="/{user}", request_method="PUT",
+        permission="user_create")
     def user_create(self):
         username = self.context.username
         request = self.request
