@@ -15,7 +15,7 @@ from devpi_common.archive import Archive, zip_dict
 from devpi_common.viewhelp import ViewLinkStore
 
 import devpi_server.views
-from devpi_server.views import tween_keyfs_transaction
+from devpi_server.views import tween_keyfs_transaction, make_uuid_headers
 
 from .functional import TestUserThings, TestIndexThings  # noqa
 
@@ -42,13 +42,23 @@ def test_invalid_username(caplog, testapp, user, status):
     assert r.status_code == code
     if status == 'warn':
         msg = "username '%s' will be invalid with next release, use characters, numbers, underscore, dash and dots only" % user
-        logmsg, = caplog.getrecords('invalid')
+        logmsg = caplog.getrecords('invalid')[-1]
         assert logmsg.message.endswith(msg)
     if status == 'fatal':
         msg = "username '%s' is invalid, use characters, numbers, underscore, dash and dots only" % user
         assert r.json['message'] == msg
 
 
+@pytest.mark.parametrize("nodeinfo,expected", [
+    ({}, (None, None)),
+    ({"uuid": "123", "role":"master"}, ("123", "123")),
+    ({"uuid": "123", "role":"replica"}, ("123", "")),
+    ({"uuid": "123", "master-uuid": "456", "role":"replica"}, ("123", "456")),
+])
+def test_make_uuid_headers(nodeinfo, expected):
+    output = make_uuid_headers(nodeinfo)
+    assert output == expected
+
 def test_simple_project(pypistage, testapp):
     name = "qpwoei"
     r = testapp.get("/root/pypi/+simple/" + name)
@@ -137,6 +147,15 @@ def test_simple_refresh(mapp, model, pypistage, testapp):
         info = pypistage._load_project_cache("hello")
     assert info == {}
 
+def test_inheritance_versiondata(mapp, model):
+    api1 = mapp.create_and_use()
+    mapp.upload_file_pypi("package-1.0.tar.gz", b'123',
+                          "package", "1.0", indexname=api1.stagename)
+    api2 = mapp.create_and_use(indexconfig={"bases": (api1.stagename,)})
+    r = mapp.getjson(api2.index + "/package")
+    assert len(r["result"]) == 1
+
+
 def test_simple_refresh_inherited(mapp, model, pypistage, testapp):
     pypistage.mock_simple("pkg", "<html/>")
     api = mapp.create_and_use()
@@ -209,6 +228,22 @@ def test_apiconfig(testapp):
     assert r.status_code == 200
     assert not "pypisubmit" in r.json["result"]
 
+class TestStatus:
+    def test_status_master(self, testapp):
+        r = testapp.get_json("/+status", status=200)
+        assert r.status_code == 200
+        data = r.json["result"]
+        assert data["role"] == "MASTER"
+
+    def test_status_replica(self, maketestapp, replica_xom):
+        testapp = maketestapp(replica_xom)
+        r = testapp.get_json("/+status", status=200)
+        assert r.status_code == 200
+        data = r.json["result"]
+        assert data["role"] == "REPLICA"
+        assert data["serial"] == replica_xom.keyfs.get_current_serial()
+
+
 def test_apiconfig_with_outside_url(testapp):
     testapp.xom.config.args.outside_url = u = "http://outside.com"
     r = testapp.get_json("/root/pypi/+api")
@@ -388,6 +423,88 @@ class TestSubmitValidation:
         path2, = mapp.get_release_paths("Pkg5")
         testapp.xget(404, path1)
         testapp.xget(200, path2)
+        r = testapp.xget(200, "%s/Pkg5/2.6" % mapp.api.index)
+        link, = r.json['result']['+links']
+        log1, log2 = link['log']
+        assert sorted(log1.keys()) == ['count', 'what', 'when', 'who']
+        assert log1['what'] == 'overwrite'
+        assert log1['who'] is None
+        assert log1['count'] == 1
+        assert sorted(log2.keys()) == ['dst', 'what', 'when', 'who']
+        assert log2['what'] == 'upload'
+        assert log2['who'] == 'user'
+        assert log2['dst'] == 'user/dev'
+
+    def test_upload_thrice_and_push(self, submit, testapp, mapp):
+        metadata = {"name": "Pkg5", "version": "2.6", ":action": "submit"}
+        submit.metadata(metadata, code=200)
+        submit.file("pkg5-2.6.tgz", b"123", {"name": "Pkg5"}, code=200)
+        submit.file("pkg5-2.6.tgz", b"1234", {"name": "Pkg5"}, code=200)
+        submit.file("pkg5-2.6.tgz", b"12345", {"name": "Pkg5"}, code=200)
+        r = testapp.xget(200, "%s/Pkg5/2.6" % mapp.api.index)
+        link, = r.json['result']['+links']
+        log1, log2 = link['log']
+        assert sorted(log1.keys()) == ['count', 'what', 'when', 'who']
+        assert log1['what'] == 'overwrite'
+        assert log1['who'] is None
+        assert log1['count'] == 2
+        assert sorted(log2.keys()) == ['dst', 'what', 'when', 'who']
+        assert log2['what'] == 'upload'
+        assert log2['who'] == 'user'
+        assert log2['dst'] == 'user/dev'
+        old_stage = mapp.api.stagename
+        mapp.create_index('prod')
+        new_stage = mapp.api.stagename
+        mapp.use(old_stage)
+        req = dict(name="Pkg5", version="2.6", targetindex=new_stage)
+        r = testapp.push("/%s" % old_stage, json.dumps(req))
+        r = testapp.xget(200, "/%s/Pkg5/2.6" % new_stage)
+        link, = r.json['result']['+links']
+        # the overwrite info should be gone
+        log1, log2 = link['log']
+        assert sorted(log1.keys()) == ['dst', 'what', 'when', 'who']
+        assert log1['what'] == 'upload'
+        assert log1['who'] == 'user'
+        assert log1['dst'] == 'user/dev'
+        assert sorted(log2.keys()) == ['dst', 'src', 'what', 'when', 'who']
+        assert log2['what'] == 'push'
+        assert log2['who'] == 'user'
+        assert log2['dst'] == 'user/prod'
+        assert log2['src'] == 'user/dev'
+
+    def test_last_modified_preserved_on_push(self, submit, testapp, mapp):
+        import time
+        metadata = {"name": "Pkg5", "version": "2.6", ":action": "submit"}
+        submit.metadata(metadata, code=200)
+        submit.file("pkg5-2.6.tgz", b"1234", {"name": "Pkg5"}, code=200)
+        old_stagename = mapp.api.stagename
+        mapp.create_index('prod')
+        new_stagename = mapp.api.stagename
+        mapp.use(old_stagename)
+        req = dict(name="Pkg5", version="2.6", targetindex=new_stagename)
+        time.sleep(1.5)  # needed to test last_modified below
+        testapp.push("/%s" % old_stagename, json.dumps(req))
+        with mapp.xom.model.keyfs.transaction(write=False):
+            old_stage = mapp.xom.model.getstage(old_stagename)
+            new_stage = mapp.xom.model.getstage(new_stagename)
+            old_entry = old_stage.get_releaselinks('Pkg5')[0].entry
+            new_entry = new_stage.get_releaselinks('Pkg5')[0].entry
+            assert old_entry.last_modified == new_entry.last_modified
+
+    def test_pypiaction_not_in_verdata_after_push(self, submit, testapp, mapp):
+        metadata = {"name": "Pkg5", "version": "2.6", ":action": "submit"}
+        submit.metadata(metadata, code=200)
+        submit.file("pkg5-2.6.tgz", b"1234", {"name": "Pkg5"}, code=200)
+        old_stagename = mapp.api.stagename
+        mapp.create_index('prod')
+        new_stagename = mapp.api.stagename
+        mapp.use(old_stagename)
+        req = dict(name="Pkg5", version="2.6", targetindex=new_stagename)
+        testapp.push("/%s" % old_stagename, json.dumps(req))
+        with mapp.xom.model.keyfs.transaction(write=False):
+            new_stage = mapp.xom.model.getstage(new_stagename)
+            verdata = new_stage.get_versiondata('Pkg5', '2.6')
+            assert ':action' not in list(verdata.keys())
 
     def test_upload_with_metadata(self, submit, testapp, mapp, pypistage):
         pypistage.mock_simple("package", '<a href="/package-1.0.zip" />')
@@ -458,6 +575,11 @@ def test_push_from_base_error(mapp, testapp, monkeypatch, pypistage):
     assert r.status_code == 400
     assert "no files for" in r.json["message"]
 
+def test_upload_docs_without_registration(mapp, testapp, monkeypatch):
+    mapp.create_and_use()
+    mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6")
+    mapp.upload_doc("pkg1-2.7.doc.zip", b'', "pkg1", "2.7", code=400)
+
 @proj
 def test_upload_and_push_internal(mapp, testapp, monkeypatch, proj):
     mapp.create_user("user1", "1")
@@ -479,10 +601,28 @@ def test_upload_and_push_internal(mapp, testapp, monkeypatch, proj):
     vv = get_view_version_links(testapp, "/user2/prod", "pkg1", "2.6",
                                 proj=proj)
     link = vv.get_link(rel="releasefile")
+    history_log = link.log
+    assert len(history_log) == 2
+    assert history_log[0]['what'] == 'upload'
+    assert history_log[0]['who'] == 'user1'
+    assert history_log[0]['dst'] == 'user1/dev'
+    assert history_log[1]['what'] == 'push'
+    assert history_log[1]['who'] == 'user1'
+    assert history_log[1]['src'] == 'user1/dev'
+    assert history_log[1]['dst'] == 'user2/prod'
     assert link.href.endswith("/pkg1-2.6.tgz")
     # we check here that the upload of docs without version was
     # automatically tied to the newest release metadata
     link = vv.get_link(rel="doczip")
+    history_log = link.log
+    assert len(history_log) == 2
+    assert history_log[0]['what'] == 'upload'
+    assert history_log[0]['who'] == 'user1'
+    assert history_log[0]['dst'] == 'user1/dev'
+    assert history_log[1]['what'] == 'push'
+    assert history_log[1]['who'] == 'user1'
+    assert history_log[1]['src'] == 'user1/dev'
+    assert history_log[1]['dst'] == 'user2/prod'
     assert link.href.endswith("/pkg1-2.6.doc.zip")
     r = testapp.get(link.href)
     archive = Archive(py.io.BytesIO(r.body))
@@ -519,12 +659,31 @@ def test_upload_and_push_with_toxresults(mapp, testapp):
         assert "user1/dev" not in actionlog[-1]
 
     vv = get_view_version_links(testapp, "/user1/prod", "pkg1", "2.6")
+    history_log = vv.get_link('releasefile').log
+    assert len(history_log) == 2
+    assert history_log[0]['what'] == 'upload'
+    assert history_log[0]['dst'] == 'user1/dev'
+    assert history_log[1]['what'] == 'push'
+    assert history_log[1]['who'] == 'user1'
+    assert history_log[1]['src'] == 'user1/dev'
+    assert history_log[1]['dst'] == 'user1/prod'
+
     links = vv.get_links("toxresult")
     assert len(links) == 2
     link1, link2 = links
     assert "user1/prod" in link1.href
     pkgmeta = json.loads(testapp.get(link1.href).body.decode("utf8"))
+
     assert pkgmeta == tox_result_data
+    history_log = link1.log
+    assert len(history_log) == 2
+    assert history_log[0]['what'] == 'upload'
+    assert history_log[0]['dst'] == 'user1/dev'
+    assert history_log[1]['what'] == 'push'
+    assert history_log[1]['who'] == 'user1'
+    assert history_log[1]['src'] == 'user1/dev'
+    assert history_log[1]['dst'] == 'user1/prod'
+
 
 def test_upload_and_push_external(mapp, testapp, reqmock):
     api = mapp.create_and_use()
@@ -596,6 +755,7 @@ def test_upload_and_delete_project(mapp, testapp):
     r = testapp.delete(api.index + "/pkg1/2.7")
     assert r.status_code == 200
     mapp.getjson(api.index + "/pkg1", code=404)
+    mapp.getjson(api.index + "/pkg1/2.7", code=404)
 
 def test_upload_with_acl(mapp):
     mapp.login("root")
@@ -631,13 +791,57 @@ def test_upload_anonymously(mapp):
                           set_whitelist=False)
 
 
+class TestPluginPermissions:
+    @pytest.fixture
+    def plugin(self):
+        class Plugin:
+            groups = ['plugingroup']
+            def devpiserver_auth_user(self, userdict, username, password):
+                if username == 'pluginuser':
+                    return dict(status="ok", groups=self.groups)
+                return dict(status="unknown")
+        return Plugin()
+
+    @pytest.fixture
+    def xom(self, makexom, plugin):
+        xom = makexom(plugins=[(plugin, None)])
+        return xom
+
+    def test_plugin_upload_group(self, mapp, plugin):
+        mapp.login("root")
+        mapp.create_and_use()  # new context and login
+        mapp.set_versiondata(dict(name="pkg1", version="1.0"))
+        mapp.login("pluginuser")
+        # pluginuser cannot write to index now
+        mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=403)
+        # now we change the acl
+        mapp.login("root")
+        mapp.set_acl([":plugingroup"])
+        mapp.login("pluginuser")
+        # we need to skip setting the whitelist here, because the user may only
+        # register and upload a package, but not modify the index
+        mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6",
+                              set_whitelist=False)
+        # if we remove the user from the group (and login again, as the groups
+        # are stored in the token) she can't upload anymore
+        plugin.groups = []
+        mapp.login("pluginuser")
+        mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=403)
+
+    def test_plugin_user_create_index(self, mapp):
+        mapp.login("pluginuser")
+        assert "pluginuser" not in mapp.getuserlist()
+        # user must already exist to succeed, since it doesn't, it should fail
+        mapp.create_index("pluginuser/dev", code=404)
+
+
 def test_upload_with_jenkins(mapp, reqmock):
     mapp.create_and_use()
-    mapp.set_uploadtrigger_jenkins("http://x.com/{pkgname}")
+    mapp.set_uploadtrigger_jenkins("http://x.com/{pkgname}/{pkgversion}")
     rec = reqmock.mockresponse(code=200, url=None)
     mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=200)
     assert len(rec.requests) == 1
-    assert rec.requests[0].url == "http://x.com/pkg1"
+    assert rec.requests[0].url == "http://x.com/pkg1/2.6"
     # XXX properly decode form
     #assert args[1]["data"]["Submit"] == "Build"
 
@@ -740,6 +944,10 @@ def test_upload_docs(mapp, testapp, proj):
     vv = get_view_version_links(testapp, api.index, "pkg1", "2.6", proj=proj)
     link = vv.get_link(rel="doczip")
     assert link.href.endswith("/pkg1-2.6.doc.zip")
+    assert len(link.log) == 1
+    assert link.log[0]['what'] == 'upload'
+    assert link.log[0]['who'] == 'user1'
+    assert link.log[0]['dst'] == 'user1/dev'
     r = testapp.get(link.href)
     archive = Archive(py.io.BytesIO(r.body))
     assert 'index.html' in archive.namelist()
@@ -889,3 +1097,99 @@ class TestTweenKeyfsTransaction:
         handler = tween_keyfs_transaction(wrapped_handler, {"xom": xom})
         response = handler(blank_request())
         assert response.headers.get("X-DEVPI-SERIAL") == str(cur_serial + 1)
+
+
+@pytest.mark.parametrize("restrict_modify", ["admin", ":admins"])
+class TestRestrictModify:
+    logins = [("root",), ("regular", "regular"), ("hello", "password")]
+
+    @pytest.fixture
+    def plugin(self):
+        class Plugin:
+            def devpiserver_auth_user(self, userdict, username, password):
+                if username == "regular" and password == "regular":
+                    return dict(status="ok", groups=["regulars"])
+                if username == "admin" and password == "admin":
+                    return dict(status="ok", groups=["admins"])
+                return dict(status="unknown")
+        return Plugin()
+
+    @pytest.fixture
+    def xom(self, makexom, plugin, restrict_modify):
+        xom = makexom(plugins=[(plugin, None)])
+        xom.config.args.restrict_modify = restrict_modify
+        return xom
+
+    def test_create_new_user(self, mapp):
+        mapp.create_user("hello", "password", code=403)
+        mapp.login("root")
+        mapp.create_user("hello", "password", code=403)
+        mapp.login("regular", "regular")
+        mapp.create_user("hello", "password", code=403)
+        mapp.login("admin", "admin")
+        assert "hello" not in mapp.getuserlist()
+        mapp.create_user("hello", "password")
+        assert "hello" in mapp.getuserlist()
+
+    def test_modify_user(self, mapp):
+        mapp.login("admin", "admin")
+        mapp.create_user("hello", "password")
+        assert "hello" in mapp.getuserlist()
+        for login in self.logins:
+            mapp.login(*login)
+            mapp.modify_user("hello", email="whatever", code=403)
+        mapp.login("admin", "admin")
+        res = mapp.getjson("/hello")["result"]
+        assert res["email"] == "hello@example.com"
+        mapp.modify_user("hello", email="whatever")
+        res = mapp.getjson("/hello")["result"]
+        assert res["email"] == "whatever"
+
+    def test_delete_user(self, mapp):
+        mapp.login("admin", "admin")
+        mapp.create_user("hello", "password")
+        for login in self.logins:
+            mapp.login(*login)
+            mapp.delete_user("hello", code=403)
+        mapp.login("admin", "admin")
+        assert "hello" in mapp.getuserlist()
+        mapp.delete_user("hello")
+        assert "hello" not in mapp.getuserlist()
+
+    def test_create_new_index(self, mapp):
+        mapp.login("admin", "admin")
+        mapp.create_user("hello", "password")
+        for login in self.logins:
+            mapp.login(*login)
+            mapp.create_index("hello/dev", code=403)
+        mapp.login("admin", "admin")
+        assert "hello/dev" not in mapp.getindexlist("hello")
+        mapp.create_index("hello/dev")
+        assert "hello/dev" in mapp.getindexlist("hello")
+
+    def test_modify_index(self, mapp):
+        mapp.login("admin", "admin")
+        mapp.create_user("hello", "password")
+        mapp.create_index("hello/dev")
+        for login in self.logins:
+            mapp.login(*login)
+            res = mapp.getjson("/hello/dev")["result"]
+            mapp.modify_index("hello/dev", res, code=403)
+        mapp.login("admin", "admin")
+        assert res["volatile"] is True
+        res["volatile"] = False
+        mapp.modify_index("hello/dev", res)
+        res = mapp.getjson("/hello/dev")["result"]
+        assert res["volatile"] is False
+
+    def test_delete_index(self, mapp):
+        mapp.login("admin", "admin")
+        mapp.create_user("hello", "password")
+        mapp.create_index("hello/dev")
+        for login in self.logins:
+            mapp.login(*login)
+            mapp.delete_index("hello/dev", code=403)
+        mapp.login("admin", "admin")
+        assert "hello/dev" in mapp.getindexlist("hello")
+        mapp.delete_index("hello/dev")
+        assert "hello/dev" not in mapp.getindexlist("hello")
